# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import pytest
from typing import Generator
from moto import mock_aws
import boto3

from instana.singletons import tracer, agent
from tests.helpers import get_first_span_by_filter

pwd = os.path.dirname(os.path.abspath(__file__))
upload_filename = os.path.abspath(pwd + "/../../data/boto3/test_upload_file.jpg")
download_target_filename = os.path.abspath(
    pwd + "/../../data/boto3/download_target_file.asdf"
)


class TestS3:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Setup and Teardown"""
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        self.mock = mock_aws()
        self.mock.start()
        self.s3 = boto3.client("s3", region_name="us-east-1")
        yield
        # Stop Moto after each test
        self.mock.stop()
        agent.options.allow_exit_as_root = False

    def test_vanilla_create_bucket(self) -> None:
        self.s3.create_bucket(Bucket="aws_bucket_name")

        result = self.s3.list_buckets()
        assert len(result["Buckets"]) == 1
        assert result["Buckets"][0]["Name"] == "aws_bucket_name"

    def test_s3_create_bucket(self) -> None:
        with tracer.start_as_current_span("test"):
            self.s3.create_bucket(Bucket="aws_bucket_name")

        result = self.s3.list_buckets()
        assert len(result["Buckets"]) == 1
        assert result["Buckets"][0]["Name"] == "aws_bucket_name"

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "CreateBucket"
        assert boto_span.data["boto3"]["ep"] == "https://s3.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"
        assert boto_span.data["boto3"]["payload"] == {"Bucket": "aws_bucket_name"}
        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"] == "https://s3.amazonaws.com:443/CreateBucket"
        )

    def test_s3_create_bucket_as_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        self.s3.create_bucket(Bucket="aws_bucket_name")

        agent.options.allow_exit_as_root = False
        result = self.s3.list_buckets()
        assert len(result["Buckets"]) == 1
        assert result["Buckets"][0]["Name"] == "aws_bucket_name"

        spans = self.recorder.queued_spans()
        assert len(spans) == 1
        boto_span = spans[0]
        assert boto_span
        assert boto_span.n == "boto3"
        assert not boto_span.p
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "CreateBucket"
        assert boto_span.data["boto3"]["ep"] == "https://s3.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"
        assert boto_span.data["boto3"]["payload"] == {"Bucket": "aws_bucket_name"}
        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"] == "https://s3.amazonaws.com:443/CreateBucket"
        )

    def test_s3_list_buckets(self) -> None:
        with tracer.start_as_current_span("test"):
            result = self.s3.list_buckets()

        assert len(result["Buckets"]) == 0
        assert result["ResponseMetadata"]["HTTPStatusCode"] == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "ListBuckets"
        assert boto_span.data["boto3"]["ep"] == "https://s3.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"
        assert boto_span.data["boto3"]["payload"] == {}
        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"] == "https://s3.amazonaws.com:443/ListBuckets"
        )

    def test_s3_vanilla_upload_file(self) -> None:
        object_name = "aws_key_name"
        bucket_name = "aws_bucket_name"

        self.s3.create_bucket(Bucket=bucket_name)
        result = self.s3.upload_file(upload_filename, bucket_name, object_name)
        assert not result

    def test_s3_upload_file(self) -> None:
        object_name = "aws_key_name"
        bucket_name = "aws_bucket_name"

        self.s3.create_bucket(Bucket=bucket_name)

        with tracer.start_as_current_span("test"):
            self.s3.upload_file(upload_filename, bucket_name, object_name)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "upload_file"
        assert boto_span.data["boto3"]["ep"] == "https://s3.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"
        payload = {
            "Filename": upload_filename,
            "Bucket": "aws_bucket_name",
            "Key": "aws_key_name",
        }
        assert boto_span.data["boto3"]["payload"] == payload
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"] == "https://s3.amazonaws.com:443/upload_file"
        )

    def test_s3_upload_file_obj(self) -> None:
        object_name = "aws_key_name"
        bucket_name = "aws_bucket_name"

        self.s3.create_bucket(Bucket=bucket_name)

        with tracer.start_as_current_span("test"):
            with open(upload_filename, "rb") as fd:
                self.s3.upload_fileobj(fd, bucket_name, object_name)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "upload_fileobj"
        assert boto_span.data["boto3"]["ep"] == "https://s3.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"
        payload = {"Bucket": "aws_bucket_name", "Key": "aws_key_name"}
        assert boto_span.data["boto3"]["payload"] == payload
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"]
            == "https://s3.amazonaws.com:443/upload_fileobj"
        )

    def test_s3_download_file(self) -> None:
        object_name = "aws_key_name"
        bucket_name = "aws_bucket_name"

        self.s3.create_bucket(Bucket=bucket_name)
        self.s3.upload_file(upload_filename, bucket_name, object_name)

        with tracer.start_as_current_span("test"):
            self.s3.download_file(bucket_name, object_name, download_target_filename)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "download_file"
        assert boto_span.data["boto3"]["ep"] == "https://s3.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"
        payload = {
            "Bucket": "aws_bucket_name",
            "Key": "aws_key_name",
            "Filename": "%s" % download_target_filename,
        }
        assert boto_span.data["boto3"]["payload"] == payload
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"]
            == "https://s3.amazonaws.com:443/download_file"
        )

    def test_s3_download_file_obj(self) -> None:
        object_name = "aws_key_name"
        bucket_name = "aws_bucket_name"

        self.s3.create_bucket(Bucket=bucket_name)
        self.s3.upload_file(upload_filename, bucket_name, object_name)

        with tracer.start_as_current_span("test"):
            with open(download_target_filename, "wb") as fd:
                self.s3.download_fileobj(bucket_name, object_name, fd)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "download_fileobj"
        assert boto_span.data["boto3"]["ep"] == "https://s3.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"]
            == "https://s3.amazonaws.com:443/download_fileobj"
        )

    def test_request_header_capture_before_call(self) -> None:
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This", "X-Capture-That"]

        # Access the event system on the S3 client
        event_system = self.s3.meta.events

        request_headers = {"X-Capture-This": "this", "X-Capture-That": "that"}

        # Create a function that adds custom headers
        def add_custom_header_before_call(params, **kwargs):
            params["headers"].update(request_headers)

        # Register the function to before-call event.
        event_system.register(
            "before-call.s3.CreateBucket", add_custom_header_before_call
        )

        with tracer.start_as_current_span("test"):
            self.s3.create_bucket(Bucket="aws_bucket_name")

        result = self.s3.list_buckets()
        assert len(result["Buckets"]) == 1
        assert result["Buckets"][0]["Name"] == "aws_bucket_name"

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "CreateBucket"
        assert boto_span.data["boto3"]["ep"] == "https://s3.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"
        assert boto_span.data["boto3"]["payload"] == {"Bucket": "aws_bucket_name"}
        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"] == "https://s3.amazonaws.com:443/CreateBucket"
        )

        assert "X-Capture-This" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-This"] == "this"
        assert "X-Capture-That" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-That"] == "that"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_request_header_capture_before_sign(self) -> None:
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Custom-1", "X-Custom-2"]

        # Access the event system on the S3 client
        event_system = self.s3.meta.events

        request_headers = {"X-Custom-1": "Value1", "X-Custom-2": "Value2"}

        # Create a function that adds custom headers
        def add_custom_header_before_sign(request, **kwargs):
            for name, value in request_headers.items():
                request.headers.add_header(name, value)

        # Register the function to before-sign event.
        event_system.register_first(
            "before-sign.s3.CreateBucket", add_custom_header_before_sign
        )

        with tracer.start_as_current_span("test"):
            self.s3.create_bucket(Bucket="aws_bucket_name")

        result = self.s3.list_buckets()
        assert len(result["Buckets"]) == 1
        assert result["Buckets"][0]["Name"] == "aws_bucket_name"

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "CreateBucket"
        assert boto_span.data["boto3"]["ep"] == "https://s3.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"
        assert boto_span.data["boto3"]["payload"] == {"Bucket": "aws_bucket_name"}
        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"] == "https://s3.amazonaws.com:443/CreateBucket"
        )

        assert "X-Custom-1" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Custom-1"] == "Value1"
        assert "X-Custom-2" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Custom-2"] == "Value2"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_response_header_capture(self) -> None:
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

        # Access the event system on the S3 client
        event_system = self.s3.meta.events

        response_headers = {
            "X-Capture-This-Too": "this too",
            "X-Capture-That-Too": "that too",
        }

        # Create a function that sets the custom headers in the after-call event.
        def modify_after_call_args(parsed, **kwargs):
            parsed["ResponseMetadata"]["HTTPHeaders"].update(response_headers)

        # Register the function to an event
        event_system.register("after-call.s3.CreateBucket", modify_after_call_args)

        with tracer.start_as_current_span("test"):
            self.s3.create_bucket(Bucket="aws_bucket_name")

        result = self.s3.list_buckets()
        assert len(result["Buckets"]) == 1
        assert result["Buckets"][0]["Name"] == "aws_bucket_name"

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "CreateBucket"
        assert boto_span.data["boto3"]["ep"] == "https://s3.amazonaws.com"
        assert boto_span.data["boto3"]["reg"] == "us-east-1"
        assert boto_span.data["boto3"]["payload"] == {"Bucket": "aws_bucket_name"}
        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert (
            boto_span.data["http"]["url"] == "https://s3.amazonaws.com:443/CreateBucket"
        )

        assert "X-Capture-This-Too" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-This-Too"] == "this too"
        assert "X-Capture-That-Too" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-That-Too"] == "that too"

        agent.options.extra_http_headers = original_extra_http_headers



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

import pytest
from typing import Generator
from sanic_testing.testing import SanicTestClient

from instana.singletons import tracer, agent
from instana.util.ids import hex_id
from tests.helpers import get_first_span_by_filter
from tests.test_utils import _TraceContextMixin
from tests.apps.sanic_app.server import app


class TestSanic(_TraceContextMixin):
    @classmethod
    def setup_class(cls) -> None:
        cls.client = SanicTestClient(app, port=1337, host="127.0.0.1")

        # Hack together a manual custom headers list; We'll use this in tests
        agent.options.extra_http_headers = [
            "X-Capture-This",
            "X-Capture-That",
            "X-Capture-This-Too",
            "X-Capture-That-Too",
        ]

    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Setup and Teardown"""
        # setup
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

    def test_vanilla_get(self) -> None:
        request, response = self.client.get("/")

        assert response.status_code == 200
        assert "X-INSTANA-T" in response.headers
        assert "X-INSTANA-S" in response.headers
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        spans = self.recorder.queued_spans()
        assert len(spans) == 1
        assert spans[0].n == "asgi"

    def test_basic_get(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As SanicTestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the sanic server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            request, response = self.client.get("/", headers=headers)

        assert response.status_code == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        span_filter = (
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        self.assertTraceContextPropagated(test_span, asgi_span)

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "127.0.0.1:1337"
        assert asgi_span.data["http"]["path"] == "/"
        assert asgi_span.data["http"]["path_tpl"] == "/"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200
        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

    def test_404(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As SanicTestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the sanic server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            request, response = self.client.get("/foo/not_an_int", headers=headers)

        assert response.status_code == 404

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        span_filter = (
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        self.assertTraceContextPropagated(test_span, asgi_span)

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "127.0.0.1:1337"
        assert asgi_span.data["http"]["path"] == "/foo/not_an_int"
        assert not asgi_span.data["http"]["path_tpl"]
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 404
        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

    def test_sanic_exception(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As SanicTestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the sanic server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            request, response = self.client.get("/wrong", headers=headers)

        assert response.status_code == 400

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        span_filter = (
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        self.assertTraceContextPropagated(test_span, asgi_span)

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "127.0.0.1:1337"
        assert asgi_span.data["http"]["path"] == "/wrong"
        assert asgi_span.data["http"]["path_tpl"] == "/wrong"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 400
        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

    def test_500_instana_exception(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As SanicTestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the sanic server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            request, response = self.client.get("/instana_exception", headers=headers)

        assert response.status_code == 500

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        span_filter = (
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        self.assertTraceContextPropagated(test_span, asgi_span)

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert asgi_span.ec == 1
        assert asgi_span.data["http"]["host"] == "127.0.0.1:1337"
        assert asgi_span.data["http"]["path"] == "/instana_exception"
        assert asgi_span.data["http"]["path_tpl"] == "/instana_exception"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 500
        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

    def test_500(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As SanicTestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the sanic server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            request, response = self.client.get("/test_request_args", headers=headers)

        assert response.status_code == 500

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        span_filter = (
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        self.assertTraceContextPropagated(test_span, asgi_span)

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert asgi_span.ec == 1
        assert asgi_span.data["http"]["host"] == "127.0.0.1:1337"
        assert asgi_span.data["http"]["path"] == "/test_request_args"
        assert asgi_span.data["http"]["path_tpl"] == "/test_request_args"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 500
        assert asgi_span.data["http"]["error"] == "Something went wrong."
        assert not asgi_span.data["http"]["params"]

    def test_path_templates(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As SanicTestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the sanic server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            request, response = self.client.get("/foo/1", headers=headers)

        assert response.status_code == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        span_filter = (
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        self.assertTraceContextPropagated(test_span, asgi_span)

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "127.0.0.1:1337"
        assert asgi_span.data["http"]["path"] == "/foo/1"
        assert asgi_span.data["http"]["path_tpl"] == "/foo/<foo_id:int>"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200
        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

    def test_secret_scrubbing(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As SanicTestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the sanic server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            request, response = self.client.get("/?secret=shhh", headers=headers)

        assert response.status_code == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        span_filter = (
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        self.assertTraceContextPropagated(test_span, asgi_span)

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "127.0.0.1:1337"
        assert asgi_span.data["http"]["path"] == "/"
        assert asgi_span.data["http"]["path_tpl"] == "/"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200
        assert not asgi_span.data["http"]["error"]
        assert asgi_span.data["http"]["params"] == "secret=<redacted>"

    def test_synthetic_request(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As SanicTestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the sanic server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
                "X-INSTANA-SYNTHETIC": "1",
            }
            request, response = self.client.get("/", headers=headers)

        assert response.status_code == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        span_filter = (
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        self.assertTraceContextPropagated(test_span, asgi_span)

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "127.0.0.1:1337"
        assert asgi_span.data["http"]["path"] == "/"
        assert asgi_span.data["http"]["path_tpl"] == "/"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200
        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

        assert asgi_span.sy
        assert not test_span.sy

    def test_request_header_capture(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As SanicTestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the sanic server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
                "X-Capture-This": "this",
                "X-Capture-That": "that",
            }
            request, response = self.client.get("/", headers=headers)

        assert response.status_code == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        span_filter = (
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        self.assertTraceContextPropagated(test_span, asgi_span)

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "127.0.0.1:1337"
        assert asgi_span.data["http"]["path"] == "/"
        assert asgi_span.data["http"]["path_tpl"] == "/"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200
        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

        assert "X-Capture-This" in asgi_span.data["http"]["header"]
        assert "this" == asgi_span.data["http"]["header"]["X-Capture-This"]
        assert "X-Capture-That" in asgi_span.data["http"]["header"]
        assert "that" == asgi_span.data["http"]["header"]["X-Capture-That"]

    def test_response_header_capture(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As SanicTestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the sanic server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            request, response = self.client.get("/response_headers", headers=headers)

        assert response.status_code == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        span_filter = (
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        self.assertTraceContextPropagated(test_span, asgi_span)

        assert "X-INSTANA-T" in response.headers
        assert response.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert "X-INSTANA-S" in response.headers
        assert response.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"
        assert "Server-Timing" in response.headers
        assert response.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "127.0.0.1:1337"
        assert asgi_span.data["http"]["path"] == "/response_headers"
        assert asgi_span.data["http"]["path_tpl"] == "/response_headers"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200

        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

        assert "X-Capture-This-Too" in asgi_span.data["http"]["header"]
        assert "this too" == asgi_span.data["http"]["header"]["X-Capture-This-Too"]
        assert "X-Capture-That-Too" in asgi_span.data["http"]["header"]
        assert "that too" == asgi_span.data["http"]["header"]["X-Capture-That-Too"]



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020


import json
import wrapt
import inspect
from typing import TYPE_CHECKING, Any, Callable, Dict, Tuple, Sequence, Type, Optional
from opentelemetry.semconv.trace import SpanAttributes

from instana.log import logger
from instana.singletons import tracer, agent
from instana.util.traceutils import get_tracer_tuple, tracing_is_off, extract_custom_headers
from instana.propagators.format import Format
from instana.span.span import get_current_span

if TYPE_CHECKING:
    from instana.span.span import InstanaSpan
    from botocore.auth import SigV4Auth
    from botocore.client import BaseClient

try:
    import boto3
    from boto3.s3 import inject

    def lambda_inject_context(payload: Dict[str, Any], span: "InstanaSpan") -> None:
        """
        When boto3 lambda client 'Invoke' is called, we want to inject the tracing context.
        boto3/botocore has specific requirements:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/lambda.html#Lambda.Client.invoke
        """
        try:
            invoke_payload = payload.get("Payload", {})

            if not isinstance(invoke_payload, dict):
                invoke_payload = json.loads(invoke_payload)

            tracer.inject(span.context, Format.HTTP_HEADERS, invoke_payload)
            payload["Payload"] = json.dumps(invoke_payload)
        except Exception:
            logger.debug("non-fatal lambda_inject_context: ", exc_info=True)

    @wrapt.patch_function_wrapper("botocore.auth", "SigV4Auth.add_auth")
    def emit_add_auth_with_instana(
        wrapped: Callable[..., None],
        instance: "SigV4Auth",
        args: Tuple[object],
        kwargs: Dict[str, Any],
    ) -> Callable[..., None]:
        current_span = get_current_span()
        if not tracing_is_off() and current_span and current_span.is_recording():
            extract_custom_headers(current_span, args[0].headers)
        return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("botocore.client", "BaseClient._make_api_call")
    def make_api_call_with_instana(
        wrapped: Callable[..., Dict[str, Any]],
        instance: Type["BaseClient"],
        arg_list: Sequence[Dict[str, Any]],
        kwargs: Dict[str, Any],
    ) -> Dict[str, Any]:
        # If we're not tracing, just return
        if tracing_is_off():
            return wrapped(*arg_list, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span("boto3", span_context=parent_context) as span:
            try:
                operation = arg_list[0]
                payload = arg_list[1]

                span.set_attribute("op", operation)
                span.set_attribute("ep", instance._endpoint.host)
                span.set_attribute("reg", instance._client_config.region_name)

                span.set_attribute(
                    SpanAttributes.HTTP_URL,
                    instance._endpoint.host + ":443/" + arg_list[0],
                )
                span.set_attribute(SpanAttributes.HTTP_METHOD, "POST")

                # Don't collect payload for SecretsManager
                if not hasattr(instance, "get_secret_value"):
                    span.set_attribute("payload", payload)

                # Inject context when invoking lambdas
                if "lambda" in instance._endpoint.host and operation == "Invoke":
                    lambda_inject_context(payload, span)

            except Exception:
                logger.debug("make_api_call_with_instana: collect error", exc_info=True)

            try:
                result = wrapped(*arg_list, **kwargs)

                if isinstance(result, dict):
                    http_dict = result.get("ResponseMetadata")
                    if isinstance(http_dict, dict):
                        status = http_dict.get("HTTPStatusCode")
                        if status is not None:
                            span.set_attribute("http.status_code", status)
                        headers = http_dict.get("HTTPHeaders")
                        extract_custom_headers(span, headers)

                return result
            except Exception as exc:
                span.mark_as_errored({"error": exc})
                raise

    def s3_inject_method_with_instana(
        wrapped: Callable[..., object],
        instance: Type["BaseClient"],
        arg_list: Sequence[object],
        kwargs: Dict[str, Any],
    ) -> Callable[..., object]:
        # If we're not tracing, just return
        if tracing_is_off():
            return wrapped(*arg_list, **kwargs)

        fas = inspect.getfullargspec(wrapped)
        fas_args = fas.args
        fas_args.remove("self")

        tracer, parent_span, _ = get_tracer_tuple()

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span("boto3", span_context=parent_context) as span:
            try:
                operation = wrapped.__name__
                span.set_attribute("op", operation)
                span.set_attribute("ep", instance._endpoint.host)
                span.set_attribute("reg", instance._client_config.region_name)

                span.set_attribute(
                    SpanAttributes.HTTP_URL,
                    instance._endpoint.host + ":443/" + operation,
                )
                span.set_attribute(SpanAttributes.HTTP_METHOD, "POST")

                arg_length = len(arg_list)
                if arg_length > 0:
                    payload = {}
                    for index in range(arg_length):
                        if fas_args[index] in ["Filename", "Bucket", "Key"]:
                            payload[fas_args[index]] = arg_list[index]
                    span.set_attribute("payload", payload)
            except Exception:
                logger.debug(
                    "s3_inject_method_with_instana: collect error", exc_info=True
                )

            try:
                return wrapped(*arg_list, **kwargs)
            except Exception as exc:
                span.mark_as_errored({"error": exc})
                raise

    for method in [
        "upload_file",
        "upload_fileobj",
        "download_file",
        "download_fileobj",
    ]:
        wrapt.wrap_function_wrapper(
            "boto3.s3.inject", method, s3_inject_method_with_instana
        )

    logger.debug("Instrumenting boto3")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2024

import unittest
from instana.util import validate_url


class TestUtil(unittest.TestCase):
    def test_validate_url(self):
        self.assertTrue(validate_url("http://localhost:3000"))
        self.assertTrue(validate_url("http://localhost:3000/"))
        self.assertTrue(validate_url("https://localhost:3000/path/item"))
        self.assertTrue(validate_url("http://localhost"))
        self.assertTrue(validate_url("https://localhost/"))
        self.assertTrue(validate_url("https://localhost/path/item"))
        self.assertTrue(validate_url("http://127.0.0.1"))
        self.assertTrue(validate_url("https://10.0.12.221/"))
        self.assertTrue(validate_url("http://[2001:db8:85a3:8d3:1319:8a2e:370:7348]/"))
        self.assertTrue(validate_url("https://[2001:db8:85a3:8d3:1319:8a2e:370:7348]:443/"))
        self.assertFalse(validate_url("boligrafo"))
        self.assertFalse(validate_url("http:boligrafo"))
        self.assertFalse(validate_url(None))



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from typing import Generator

from fastapi.testclient import TestClient
import pytest
from instana.singletons import tracer, agent

from instana.util.ids import hex_id
from tests.apps.fastapi_app.app import fastapi_server
from tests.helpers import get_first_span_by_filter


class TestFastAPI:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # We are using the TestClient from Starlette/FastAPI to make it easier.
        self.client = TestClient(fastapi_server)

        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        # Hack together a manual custom headers list; We'll use this in tests
        agent.options.extra_http_headers = [
            "X-Capture-This",
            "X-Capture-That",
            "X-Capture-This-Too",
            "X-Capture-That-Too",
        ]

    def test_vanilla_get(self) -> None:
        result = self.client.get("/")

        assert result
        assert result.status_code == 200
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        # FastAPI instrumentation (like all instrumentation) _always_ traces
        # unless told otherwise
        spans = self.recorder.queued_spans()

        assert len(spans) == 1
        assert spans[0].n == "asgi"

    def test_basic_get(self) -> None:
        result = None
        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            result = self.client.get("/", headers=headers)

        assert result
        assert result.status_code == 200
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        # TODO: after support httpx, the expected value will be 3.
        assert len(spans) == 2

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"  # noqa: E731
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        assert test_span.t == asgi_span.t
        assert test_span.s == asgi_span.p
        
        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "testserver"
        assert asgi_span.data["http"]["path"] == "/"
        assert asgi_span.data["http"]["path_tpl"] == "/"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200

        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

    def test_400(self) -> None:
        result = None
        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            result = self.client.get("/400", headers=headers)

        assert result
        assert result.status_code == 400
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        # TODO: after support httpx, the expected value will be 3.
        assert len(spans) == 2

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"  # noqa: E731
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        assert test_span.t == asgi_span.t
        assert test_span.s == asgi_span.p
        
        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "testserver"
        assert asgi_span.data["http"]["path"] == "/400"
        assert asgi_span.data["http"]["path_tpl"] == "/400"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 400

        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

    def test_500(self) -> None:
        result = None
        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            result = self.client.get("/500", headers=headers)

        assert result
        assert result.status_code == 500
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        # TODO: after support httpx, the expected value will be 3.
        assert len(spans) == 2

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"  # noqa: E731
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        assert test_span.t == asgi_span.t
        assert test_span.s == asgi_span.p
        
        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert asgi_span.ec == 1
        assert asgi_span.data["http"]["host"] == "testserver"
        assert asgi_span.data["http"]["path"] == "/500"
        assert asgi_span.data["http"]["path_tpl"] == "/500"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 500
        assert asgi_span.data["http"]["error"] == "500 response"
        assert not asgi_span.data["http"]["params"]

    def test_path_templates(self) -> None:
        result = None
        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            result = self.client.get("/users/1", headers=headers)

        assert result
        assert result.status_code == 200
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        # TODO: after support httpx, the expected value will be 3.
        assert len(spans) == 2

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"  # noqa: E731
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        assert test_span.t == asgi_span.t
        assert test_span.s == asgi_span.p
        
        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "testserver"
        assert asgi_span.data["http"]["path"] == "/users/1"
        assert asgi_span.data["http"]["path_tpl"] == "/users/{user_id}"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200
        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

    def test_secret_scrubbing(self) -> None:
        result = None
        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            result = self.client.get("/?secret=shhh", headers=headers)

        assert result
        assert result.status_code == 200
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        # TODO: after support httpx, the expected value will be 3.
        assert len(spans) == 2

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"  # noqa: E731
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        assert test_span.t == asgi_span.t
        assert test_span.s == asgi_span.p
        
        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "testserver"
        assert asgi_span.data["http"]["path"] == "/"
        assert asgi_span.data["http"]["path_tpl"] == "/"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200

        assert not asgi_span.data["http"]["error"]
        assert asgi_span.data["http"]["params"] == "secret=<redacted>"

    def test_synthetic_request(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
                "X-INSTANA-SYNTHETIC": "1",
            }
            result = self.client.get("/", headers=headers)

        assert result
        assert result.status_code == 200
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        # TODO: after support httpx, the expected value will be 3.
        assert len(spans) == 2

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"  # noqa: E731
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        assert test_span.t == asgi_span.t
        assert test_span.s == asgi_span.p
        
        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "testserver"
        assert asgi_span.data["http"]["path"] == "/"
        assert asgi_span.data["http"]["path_tpl"] == "/"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200

        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

        assert asgi_span.sy
        assert not test_span.sy

    def test_request_header_capture(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
                "X-Capture-This": "this", 
                "X-Capture-That": "that",
            }
            result = self.client.get("/", headers=headers)

        assert result
        assert result.status_code == 200
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        # TODO: after support httpx, the expected value will be 3.
        assert len(spans) == 2

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"  # noqa: E731
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        assert test_span.t == asgi_span.t
        assert test_span.s == asgi_span.p
        
        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span.s)        
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "testserver"
        assert asgi_span.data["http"]["path"] == "/"
        assert asgi_span.data["http"]["path_tpl"] == "/"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200

        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

        assert "X-Capture-This" in asgi_span.data["http"]["header"]
        assert asgi_span.data["http"]["header"]["X-Capture-This"] == "this"
        assert "X-Capture-That" in asgi_span.data["http"]["header"]
        assert asgi_span.data["http"]["header"]["X-Capture-That"] == "that"

    def test_response_header_capture(self) -> None:
        # The background FastAPI server is pre-configured with custom headers 
        # to capture.

        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            result = self.client.get("/response_headers", headers=headers)

        assert result
        assert result.status_code == 200
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        # TODO: after support httpx, the expected value will be 3.
        assert len(spans) == 2

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"  # noqa: E731
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        assert test_span.t == asgi_span.t
        assert test_span.s == asgi_span.p
        
        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "testserver"
        assert asgi_span.data["http"]["path"] == "/response_headers"
        assert asgi_span.data["http"]["path_tpl"] == "/response_headers"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200

        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]

        assert "X-Capture-This-Too" in asgi_span.data["http"]["header"]
        assert asgi_span.data["http"]["header"]["X-Capture-This-Too"] == "this too"
        assert "X-Capture-That-Too" in asgi_span.data["http"]["header"]
        assert asgi_span.data["http"]["header"]["X-Capture-That-Too"] == "that too"

    def test_non_async_simple(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            result = self.client.get("/non_async_simple", headers=headers)

        assert result
        assert result.status_code == 200
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi" and span.p == test_span.s  # noqa: E731
        asgi_span1 = get_first_span_by_filter(spans, span_filter)
        assert asgi_span1

        span_filter = lambda span: span.n == "asgi" and span.p == asgi_span1.s  # noqa: E731
        asgi_span2 = get_first_span_by_filter(spans, span_filter)
        assert asgi_span2

        # Same traceId
        traceId = test_span.t
        assert asgi_span1.t == traceId
        assert asgi_span2.t == traceId

        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span1.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span1.s)
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span1.t)}"

        assert not asgi_span1.ec
        assert asgi_span1.data["http"]["host"] == "testserver"
        assert asgi_span1.data["http"]["path"] == "/non_async_simple"
        assert asgi_span1.data["http"]["path_tpl"] == "/non_async_simple"
        assert asgi_span1.data["http"]["method"] == "GET"
        assert asgi_span1.data["http"]["status"] == 200
        assert not asgi_span1.data["http"]["error"]
        assert not asgi_span1.data["http"]["params"]

        assert not asgi_span2.ec
        assert asgi_span2.data["http"]["host"], "testserver"
        assert asgi_span2.data["http"]["path"], "/users/1"
        assert asgi_span2.data["http"]["path_tpl"], "/users/{user_id}"
        assert asgi_span2.data["http"]["method"], "GET"
        assert asgi_span2.data["http"]["status"], 200
        assert not asgi_span2.data["http"]["error"]
        assert not asgi_span2.data["http"]["params"]

    def test_non_async_threadpool(self) -> None:
        with tracer.start_as_current_span("test") as span:
            # As TestClient() is based on httpx, and we don't support it yet,
            # we must pass the SDK trace_id and span_id to the ASGI server.
            span_context = span.get_span_context()
            headers = {
                "X-INSTANA-T": hex_id(span_context.trace_id),
                "X-INSTANA-S": hex_id(span_context.span_id),
            }
            result = self.client.get("/non_async_threadpool", headers=headers)

        assert result
        assert result.status_code == 200
        assert "X-INSTANA-T" in result.headers
        assert "X-INSTANA-S" in result.headers
        assert "X-INSTANA-L" in result.headers
        assert "Server-Timing" in result.headers
        assert result.headers["X-INSTANA-L"] == "1"

        spans = self.recorder.queued_spans()
        # TODO: after support httpx, the expected value will be 3.
        assert len(spans) == 2

        span_filter = (  # noqa: E731
            lambda span: span.n == "sdk" and span.data["sdk"]["name"] == "test"
        )
        test_span = get_first_span_by_filter(spans, span_filter)
        assert test_span

        span_filter = lambda span: span.n == "asgi"  # noqa: E731
        asgi_span = get_first_span_by_filter(spans, span_filter)
        assert asgi_span

        assert test_span.t == asgi_span.t
        assert test_span.s == asgi_span.p
        
        assert result.headers["X-INSTANA-T"] == hex_id(asgi_span.t)
        assert result.headers["X-INSTANA-S"] == hex_id(asgi_span.s)
        assert result.headers["Server-Timing"] == f"intid;desc={hex_id(asgi_span.t)}"

        assert not asgi_span.ec
        assert asgi_span.data["http"]["host"] == "testserver"
        assert asgi_span.data["http"]["path"] == "/non_async_threadpool"
        assert asgi_span.data["http"]["path_tpl"] == "/non_async_threadpool"
        assert asgi_span.data["http"]["method"] == "GET"
        assert asgi_span.data["http"]["status"] == 200

        assert not asgi_span.data["http"]["error"]
        assert not asgi_span.data["http"]["params"]



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

try:
    from pyramid.httpexceptions import HTTPException
    from pyramid.path import caller_package
    from pyramid.settings import aslist
    from pyramid.tweens import EXCVIEW
    from pyramid.config import Configurator
    from typing import TYPE_CHECKING, Dict, Any, Callable, Tuple
    import wrapt

    from opentelemetry.semconv.trace import SpanAttributes
    from opentelemetry.trace import SpanKind

    from instana.log import logger
    from instana.singletons import tracer, agent
    from instana.util.secrets import strip_secrets_from_query
    from instana.util.traceutils import extract_custom_headers
    from instana.propagators.format import Format

    if TYPE_CHECKING:
        from pyramid.request import Request
        from pyramid.response import Response
        from pyramid.registry import Registry

    class InstanaTweenFactory(object):
        """A factory that provides Instana instrumentation tween for Pyramid apps"""

        def __init__(
            self, handler: Callable[["Request"], "Response"], registry: "Registry"
        ) -> None:
            self.handler = handler

        def __call__(self, request: "Request") -> "Response":
            ctx = tracer.extract(Format.HTTP_HEADERS, dict(request.headers))

            with tracer.start_as_current_span("wsgi", span_context=ctx) as span:
                span.set_attribute("span.kind", SpanKind.SERVER)
                span.set_attribute("http.host", request.host)
                span.set_attribute(SpanAttributes.HTTP_METHOD, request.method)
                span.set_attribute(SpanAttributes.HTTP_URL, request.path)

                extract_custom_headers(span, request.headers)

                if len(request.query_string):
                    scrubbed_params = strip_secrets_from_query(
                        request.query_string,
                        agent.options.secrets_matcher,
                        agent.options.secrets_list,
                    )
                    span.set_attribute("http.params", scrubbed_params)

                response = None
                try:
                    response = self.handler(request)
                    if request.matched_route is not None:
                        span.set_attribute(
                            "http.path_tpl", request.matched_route.pattern
                        )

                    extract_custom_headers(span, response.headers)

                    tracer.inject(span.context, Format.HTTP_HEADERS, response.headers)
                except HTTPException as e:
                    response = e
                    logger.debug(
                        "Pyramid InstanaTweenFactory HTTPException: ", exc_info=True
                    )
                except BaseException as e:
                    span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, 500)
                    span.record_exception(e)

                    logger.debug(
                        "Pyramid InstanaTweenFactory BaseException: ", exc_info=True
                    )
                finally:
                    if response:
                        span.set_attribute(
                            SpanAttributes.HTTP_STATUS_CODE, response.status_int
                        )

                        if 500 <= response.status_int:
                            if response.exception:
                                span.record_exception(response.exception)
                            span.assure_errored()

                return response

    INSTANA_TWEEN = __name__ + ".InstanaTweenFactory"

    # implicit tween ordering
    def includeme(config: Configurator) -> None:
        logger.debug("Instrumenting pyramid")
        config.add_tween(INSTANA_TWEEN)

    # explicit tween ordering
    @wrapt.patch_function_wrapper("pyramid.config", "Configurator.__init__")
    def init_with_instana(
        wrapped: Callable[..., Configurator.__init__],
        instance: Configurator,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ):
        settings = kwargs.get("settings", {})
        tweens = aslist(settings.get("pyramid.tweens", []))

        if tweens and INSTANA_TWEEN not in settings:
            # pyramid.tweens.EXCVIEW is the name of built-in exception view provided by
            # pyramid.  We need our tween to be before it, otherwise unhandled
            # exceptions will be caught before they reach our tween.
            if EXCVIEW in tweens:
                tweens = [INSTANA_TWEEN] + tweens
            else:
                tweens = [INSTANA_TWEEN] + tweens + [EXCVIEW]
            settings["pyramid.tweens"] = "\n".join(tweens)
            kwargs["settings"] = settings

        if not kwargs.get("package", None):
            kwargs["package"] = caller_package()

        wrapped(*args, **kwargs)
        instance.include(__name__)

except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2017

import pytest
from opentelemetry.trace.span import _SPAN_ID_MAX_VALUE, INVALID_SPAN_ID

import instana


def test_id_generation():
    count = 0
    while count <= 10000:
        id = instana.util.ids.generate_id()
        assert id >= 0
        assert id > INVALID_SPAN_ID
        assert id <= _SPAN_ID_MAX_VALUE
        count += 1


@pytest.mark.parametrize(
    "str_id, id",
    [
        ("BADCAFFE", 3135025150),
        ("abcdef", 11259375),
        ("0123456789abcdef", 81985529216486895),
        ("0x0123456789abcdef0123456789abcdef", 1512366075204170929049582354406559215),
        (None, INVALID_SPAN_ID),
        (1234, INVALID_SPAN_ID),
        ([1234], INVALID_SPAN_ID),
        ("0xZZZZZZ", INVALID_SPAN_ID),
        ("ZZZZZZ", INVALID_SPAN_ID),
        (b"BADCAFFE", 3135025150),
        (b"abcdef", 11259375),
        (b"0123456789abcdef", 81985529216486895),
        (b"0x0123456789abcdef0123456789abcdef", 1512366075204170929049582354406559215),
    ],
)
def test_header_to_long_id(str_id, id):
    result = instana.util.ids.header_to_long_id(str_id)
    assert result == id


@pytest.mark.parametrize(
    "str_id, id",
    [
        ("BADCAFFE", 3135025150),
        ("abcdef", 11259375),
        ("0123456789abcdef", 81985529216486895),
        ("0x0123456789abcdef0123456789abcdef", 81985529216486895),
        (None, INVALID_SPAN_ID),
        (1234, INVALID_SPAN_ID),
        ([1234], INVALID_SPAN_ID),
        ("0xZZZZZZ", INVALID_SPAN_ID),
        ("ZZZZZZ", INVALID_SPAN_ID),
        (b"BADCAFFE", 3135025150),
        (b"abcdef", 11259375),
        (b"0123456789abcdef", 81985529216486895),
        (b"0x0123456789abcdef0123456789abcdef", 81985529216486895),
    ],
)
def test_header_to_id(str_id, id):
    result = instana.util.ids.header_to_id(str_id)
    assert result == id



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import importlib.util
import os
import sys
from typing import Any, Dict

import pytest
from opentelemetry.context.context import Context
from opentelemetry.trace import set_span_in_context
from opentelemetry.trace.span import format_span_id

if importlib.util.find_spec("celery"):
    pytest_plugins = ("celery.contrib.pytest",)


from instana.agent.host import HostAgent
from instana.collector.base import BaseCollector
from instana.recorder import StanRecorder
from instana.span.base_span import BaseSpan
from instana.span.span import InstanaSpan
from instana.span_context import SpanContext
from instana.tracer import InstanaTracerProvider
from instana.fsm import TheMachine

collect_ignore_glob = [
    "*test_gevent*",
    "*collector/test_gcr*",
    "*agent/test_google*",
]

# # Cassandra and gevent tests are run in dedicated jobs on CircleCI and will
# # be run explicitly.  (So always exclude them here)
if not os.environ.get("CASSANDRA_TEST"):
    collect_ignore_glob.append("*test_cassandra*")

if not os.environ.get("COUCHBASE_TEST"):
    collect_ignore_glob.append("*test_couchbase*")

if not os.environ.get("GEVENT_STARLETTE_TEST"):
    collect_ignore_glob.append("*test_gevent*")
    collect_ignore_glob.append("*test_starlette*")


if sys.version_info >= (3, 13):
    # Currently not installable dependencies because of 3.13 incompatibilities
    collect_ignore_glob.append("*test_sanic*")


if sys.version_info >= (3, 14):
    # Currently not installable dependencies because of 3.14 incompatibilities
    collect_ignore_glob.append("*test_fastapi*")
    # aiohttp-server tests failing due to deprecated methods used
    collect_ignore_glob.append("*test_aiohttp_server*")


@pytest.fixture(scope="session")
def celery_config():
    return {
        "broker_connection_retry_on_startup": True,
        "broker_url": "redis://localhost:6379",
        "result_backend": "redis://localhost:6379",
    }


@pytest.fixture(scope="session")
def celery_enable_logging():
    return True


@pytest.fixture(scope="session")
def celery_includes():
    return {"tests.frameworks.test_celery"}


@pytest.fixture
def trace_id() -> int:
    return 1812338823475918251


@pytest.fixture
def span_id() -> int:
    return 6895521157646639861


@pytest.fixture
def hex_trace_id(trace_id: int) -> str:
    # Using format_span_id() to return a 16-byte hexadecimal string, instead of
    # the 32-byte hexadecimal string from format_trace_id().
    return format_span_id(trace_id)


@pytest.fixture
def hex_span_id(span_id: int) -> str:
    return format_span_id(span_id)


@pytest.fixture
def span_processor() -> StanRecorder:
    rec = StanRecorder(HostAgent())
    rec.THREAD_NAME = "InstanaSpan Recorder Test"
    return rec


@pytest.fixture
def tracer_provider(span_processor: StanRecorder) -> InstanaTracerProvider:
    return InstanaTracerProvider(span_processor=span_processor, exporter=HostAgent())


@pytest.fixture
def span_context(trace_id: int, span_id: int) -> SpanContext:
    return SpanContext(
        trace_id=trace_id,
        span_id=span_id,
        is_remote=False,
    )


@pytest.fixture
def span(span_context: SpanContext, span_processor: StanRecorder) -> InstanaSpan:
    span_name = "test-span"
    return InstanaSpan(span_name, span_context, span_processor)


@pytest.fixture
def base_span(span: InstanaSpan) -> BaseSpan:
    return BaseSpan(span, None)


@pytest.fixture
def context(span: InstanaSpan) -> Context:
    return set_span_in_context(span)


def always_true(_: object, *args: object, **kwargs: object) -> bool:
    return True


# Mocking HostAgent.can_send()
@pytest.fixture(autouse=True)
def can_send(monkeypatch, request) -> None:
    """Return always True for HostAgent.can_send()"""
    if "original" in request.keywords:
        # If using the `@pytest.mark.original` marker before the test function,
        # uses the original HostAgent.can_send()
        monkeypatch.setattr(HostAgent, "can_send", HostAgent.can_send)
    else:
        monkeypatch.setattr(HostAgent, "can_send", always_true)


# Mocking HostAgent.get_from_structure()
@pytest.fixture(autouse=True)
def get_from_structure(monkeypatch, request) -> None:
    """
    Retrieves the From data that is reported alongside monitoring data.
    @return: dict()
    """

    def _get_from_structure(_: object) -> Dict[str, Any]:
        return {"e": os.getpid(), "h": "fake"}

    if "original" in request.keywords:
        # If using the `@pytest.mark.original` marker before the test function,
        # uses the original HostAgent.get_from_structure()
        monkeypatch.setattr(
            HostAgent, "get_from_structure", HostAgent.get_from_structure
        )
    else:
        monkeypatch.setattr(HostAgent, "get_from_structure", _get_from_structure)


# Mocking BaseCollector.prepare_and_report_data()
@pytest.fixture(autouse=True)
def prepare_and_report_data(monkeypatch, request):
    """Return always True for BaseCollector.prepare_and_report_data()"""
    if "original" in request.keywords:
        # If using the `@pytest.mark.original` marker before the test function,
        # uses the original BaseCollector.prepare_and_report_data()
        monkeypatch.setattr(
            BaseCollector,
            "prepare_and_report_data",
            BaseCollector.prepare_and_report_data,
        )
    else:
        monkeypatch.setattr(BaseCollector, "prepare_and_report_data", always_true)


# Mocking HostAgent.is_agent_listening()
@pytest.fixture(autouse=True)
def is_agent_listening(monkeypatch, request) -> None:
    """Always return `True` for `HostAgent.is_agent_listening()`"""
    if "original" in request.keywords:
        # If using the `@pytest.mark.original` marker before the test function,
        # uses the original HostAgent.is_agent_listening()
        monkeypatch.setattr(
            HostAgent, "is_agent_listening", HostAgent.is_agent_listening
        )
    else:
        monkeypatch.setattr(HostAgent, "is_agent_listening", always_true)


@pytest.fixture(autouse=True)
def lookup_agent_host(monkeypatch, request) -> None:
    """Always return `True` for `TheMachine.lookup_agent_host()`"""
    if "original" in request.keywords:
        # If using the `@pytest.mark.original` marker before the test function,
        # uses the original TheMachine.lookup_agent_host()
        monkeypatch.setattr(
            TheMachine, "lookup_agent_host", TheMachine.lookup_agent_host
        )
    else:
        monkeypatch.setattr(TheMachine, "lookup_agent_host", always_true)


@pytest.fixture(autouse=True)
def announce_sensor(monkeypatch, request) -> None:
    """Always return `True` for `TheMachine.announce_sensor()`"""
    if "original" in request.keywords:
        # If using the `@pytest.mark.original` marker before the test function,
        # uses the original TheMachine.announce_sensor()
        monkeypatch.setattr(TheMachine, "announce_sensor", TheMachine.announce_sensor)
    else:
        monkeypatch.setattr(TheMachine, "announce_sensor", always_true)


@pytest.fixture(autouse=True)
def announce(monkeypatch, request) -> None:
    """Always return `True` for `Host.announce()`"""
    if "original" in request.keywords:
        # If using the `@pytest.mark.original` marker before the test function,
        # uses the original HostAgent.announce()
        monkeypatch.setattr(HostAgent, "announce", HostAgent.announce)
    else:
        monkeypatch.setattr(HostAgent, "announce", always_true)



# (c) Copyright IBM Corp. 2023

"""
The Instana agent (for AWS EKS Fargate) that manages
monitoring state and reporting that data.
"""
import os
import time
from instana.options import EKSFargateOptions
from instana.collector.aws_eks_fargate import EKSFargateCollector
from instana.collector.helpers.eks.process import get_pod_name
from instana.log import logger
from instana.util import to_json
from instana.agent.base import BaseAgent
from instana.version import VERSION


class EKSFargateAgent(BaseAgent):
    """ In-process agent for AWS Fargate """
    def __init__(self):
        super(EKSFargateAgent, self).__init__()

        self.options = EKSFargateOptions()
        self.collector = None
        self.report_headers = None
        self._can_send = False
        self.podname = get_pod_name()

        # Update log level (if INSTANA_LOG_LEVEL was set)
        self.update_log_level()

        logger.info("Stan is on the EKS Pod on AWS Fargate scene.  Starting Instana instrumentation version: %s", VERSION)

        if self._validate_options():
            self._can_send = True
            self.collector = EKSFargateCollector(self)
            self.collector.start()
        else:
            logger.warning("Required INSTANA_AGENT_KEY and/or INSTANA_ENDPOINT_URL environment variables not set.  "
                           "We will not be able to monitor this Pod.")

    def can_send(self):
        """
        Are we in a state where we can send data?
        @return: Boolean
        """
        return self._can_send

    def get_from_structure(self):
        """
        Retrieves the From data that is reported alongside monitoring data.
        @return: dict()
        """

        return {'hl': True, 'cp': 'k8s', 'e': self.podname}

    def report_data_payload(self, payload):
        """
        Used to report metrics and span data to the endpoint URL in self.options.endpoint_url
        """
        response = None
        try:
            if self.report_headers is None:
                # Prepare request headers
                self.report_headers = dict()
                self.report_headers["Content-Type"] = "application/json"
                self.report_headers["X-Instana-Host"] = self.podname
                self.report_headers["X-Instana-Key"] = self.options.agent_key

            response = self.client.post(self.__data_bundle_url(),
                                        data=to_json(payload),
                                        headers=self.report_headers,
                                        timeout=self.options.timeout,
                                        verify=self.options.ssl_verify,
                                        proxies=self.options.endpoint_proxy)

            if not 200 <= response.status_code < 300:
                logger.info("report_data_payload: Instana responded with status code %s", response.status_code)
        except Exception as exc:
            logger.debug("report_data_payload: connection error (%s)", type(exc))
        return response

    def _validate_options(self):
        """
        Validate that the options used by this Agent are valid.  e.g. can we report data?
        """
        return self.options.endpoint_url is not None and self.options.agent_key is not None

    def __data_bundle_url(self):
        """
        URL for posting metrics to the host agent.  Only valid when announced.
        """
        return "%s/bundle" % self.options.endpoint_url



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import signal
import sys
import threading
from typing import TYPE_CHECKING, List, Optional, Tuple

from instana.autoprofile.profile import CallSite, Profile
from instana.autoprofile.runtime import RuntimeInfo
from instana.log import logger

if TYPE_CHECKING:
    from types import FrameType


if RuntimeInfo.GEVENT:
    import gevent


class BlockSampler(object):
    SAMPLING_RATE = 0.05
    MAX_TRACEBACK_SIZE = 25  # number of frames

    def __init__(self, profiler: Profile) -> None:
        self.profiler = profiler
        self.ready = False
        self.top = None
        self.top_lock = threading.Lock()
        self.prev_signal_handler = None
        self.sampler_active = False

    def setup(self) -> None:
        if self.profiler.get_option("block_sampler_disabled"):
            return

        if not RuntimeInfo.OS_LINUX and not RuntimeInfo.OS_DARWIN:
            logger.debug("CPU profiler is only supported on Linux and OS X.")
            return

        sample_time = self.SAMPLING_RATE * 1000

        main_thread_id = (
            gevent._threading.get_ident()
            if RuntimeInfo.GEVENT
            else threading.current_thread().ident
        )

        def _sample(signum: object, signal_frame: "FrameType") -> None:
            if self.sampler_active:
                return
            self.sampler_active = True

            with self.top_lock:
                try:
                    self.process_sample(signal_frame, sample_time, main_thread_id)
                    signal_frame = None
                except Exception:
                    logger.error("Error processing sample", exc_info=True)

            self.sampler_active = False

        self.prev_signal_handler = signal.signal(signal.SIGALRM, _sample)

        self.ready = True

    def destroy(self) -> None:
        if not self.ready:
            return

        signal.signal(signal.SIGALRM, self.prev_signal_handler)

    def reset(self) -> None:
        self.top = CallSite("", "", 0)

    def start_sampler(self) -> None:
        logger.debug("Activating block sampler.")

        signal.setitimer(signal.ITIMER_REAL, self.SAMPLING_RATE, self.SAMPLING_RATE)

    def stop_sampler(self) -> None:
        signal.setitimer(signal.ITIMER_REAL, 0)

        logger.debug("Deactivating block sampler.")

    def build_profile(self, duration: int, timespan: int) -> Profile:
        with self.top_lock:
            self.top.normalize(duration)
            self.top.floor()

            profile = Profile(
                Profile.CATEGORY_TIME,
                Profile.TYPE_BLOCKING_CALLS,
                Profile.UNIT_MILLISECOND,
                self.top.children.values(),
                duration,
                timespan,
            )

            return profile

    def process_sample(
        self, signal_frame: "FrameType", sample_time: int, main_thread_id: int
    ) -> None:
        if self.top:
            current_frames = sys._current_frames()
            items = current_frames.items()
            for thread_id, thread_frame in items:
                if thread_id == main_thread_id:
                    thread_frame = signal_frame

                stack = self.recover_stack(thread_frame)
                if stack:
                    current_node = self.top
                    for func_name, filename, lineno in reversed(stack):
                        current_node = current_node.find_or_add_child(
                            func_name, filename, lineno
                        )
                    current_node.increment(sample_time, 1)

                thread_id, thread_frame, stack = None, None, None

            items = None
            current_frames = None

    def recover_stack(
        self, thread_frame: "FrameType"
    ) -> Optional[List[Tuple[str, str, int]]]:
        stack = []

        depth = 0
        while thread_frame is not None and depth <= self.MAX_TRACEBACK_SIZE:
            if (
                thread_frame.f_code
                and thread_frame.f_code.co_name
                and thread_frame.f_code.co_filename
            ):
                func_name = thread_frame.f_code.co_name
                filename = thread_frame.f_code.co_filename
                lineno = thread_frame.f_lineno

                if filename and self.profiler.frame_cache.is_profiler_frame(filename):
                    return None

                stack.append((func_name, filename, lineno))

                thread_frame = thread_frame.f_back

            depth += 1

        if len(stack) == 0:
            return None
        else:
            return stack



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""Collection helper for the process"""

import grp
import os
import pwd

from instana.collector.helpers.base import BaseHelper
from instana.log import logger
from instana.util import DictionaryOfStan
from instana.util.runtime import get_proc_cmdline
from instana.util.secrets import contains_secret


class ProcessHelper(BaseHelper):
    """Helper class to collect metrics for this process"""

    def collect_metrics(self, **kwargs):
        plugin_data = dict()
        try:
            plugin_data["name"] = "com.instana.plugin.process"
            plugin_data["entityId"] = str(os.getpid())
            plugin_data["data"] = DictionaryOfStan()
            plugin_data["data"]["pid"] = int(os.getpid())

            if kwargs.get("with_snapshot"):
                self._collect_process_snapshot(plugin_data)
        except Exception:
            logger.debug("ProcessHelper.collect_metrics: ", exc_info=True)
        return plugin_data

    def _collect_process_snapshot(self, plugin_data):
        try:
            env = dict()
            for key in os.environ:
                if contains_secret(
                    key,
                    self.collector.agent.options.secrets_matcher,
                    self.collector.agent.options.secrets_list,
                ):
                    env[key] = "<ignored>"
                else:
                    env[key] = os.environ[key]
            plugin_data["data"]["env"] = env
            if os.path.isfile("/proc/self/exe"):
                plugin_data["data"]["exec"] = os.readlink("/proc/self/exe")
            else:
                logger.debug("Can't access /proc/self/exe...")

            cmdline = get_proc_cmdline()
            if len(cmdline) > 1:
                # drop the exe
                cmdline.pop(0)
            plugin_data["data"]["args"] = cmdline
            try:
                euid = os.geteuid()
                egid = os.getegid()
                plugin_data["data"]["user"] = pwd.getpwuid(euid)
                plugin_data["data"]["group"] = grp.getgrgid(egid).gr_name
            except Exception:
                logger.debug("euid/egid detection: ", exc_info=True)

            plugin_data["data"]["start"] = self.collector.fetching_start_time

        except Exception:
            logger.debug("ProcessHelper._collect_process_snapshot: ", exc_info=True)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021


from typing import TYPE_CHECKING, Any, Callable, Dict, Tuple

import wrapt

from instana.log import logger
from instana.propagators.format import Format
from instana.singletons import tracer
from instana.util.traceutils import get_tracer_tuple, tracing_is_off

if TYPE_CHECKING:
    from instana.span.span import InstanaSpan

try:
    from google.cloud import pubsub_v1

    def _set_publisher_attributes(
        span: "InstanaSpan",
        topic_path: str,
    ) -> None:
        span.set_attribute("gcps.op", "publish")
        # Fully qualified identifier is in the form of
        # `projects/{project_id}/topic/{topic_name}`
        project_id, topic_name = topic_path.split("/")[1::2]
        span.set_attribute("gcps.projid", project_id)
        span.set_attribute("gcps.top", topic_name)

    def _set_consumer_attributes(
        span: "InstanaSpan",
        subscription_path: str,
    ) -> None:
        span.set_attribute("gcps.op", "consume")
        # Fully qualified identifier is in the form of
        # `projects/{project_id}/subscriptions/{subscription_name}`
        project_id, subscription_id = subscription_path.split("/")[1::2]
        span.set_attribute("gcps.projid", project_id)
        span.set_attribute("gcps.sub", subscription_id)

    @wrapt.patch_function_wrapper("google.cloud.pubsub_v1", "PublisherClient.publish")
    def publish_with_instana(
        wrapped: Callable[..., object],
        instance: pubsub_v1.PublisherClient,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        """References:
        - PublisherClient.publish(topic_path, messages, metadata)
        """
        # return early if we're not tracing
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "gcps-producer", span_context=parent_context
        ) as span:
            # trace continuity, inject to the span context
            headers = {}
            tracer.inject(
                span.context,
                Format.TEXT_MAP,
                headers,
                disable_w3c_trace_context=True,
            )

            headers = {key: str(value) for key, value in headers.items()}

            # update the metadata dict with instana trace attributes
            kwargs.update(headers)

            _set_publisher_attributes(span, topic_path=args[0])

            try:
                rv = wrapped(*args, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return rv

    @wrapt.patch_function_wrapper(
        "google.cloud.pubsub_v1", "SubscriberClient.subscribe"
    )
    def subscribe_with_instana(
        wrapped: Callable[..., object],
        instance: pubsub_v1.SubscriberClient,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        """References:
        - SubscriberClient.subscribe(subscription_path, callback)
        - callback(message) is called from the subscription future
        """

        def callback_with_instana(message):
            if message.attributes:
                parent_context = tracer.extract(
                    Format.TEXT_MAP, message.attributes, disable_w3c_trace_context=True
                )
            else:
                parent_context = None

            with tracer.start_as_current_span(
                "gcps-consumer", span_context=parent_context
            ) as span:
                _set_consumer_attributes(span, subscription_path=args[0])
                try:
                    callback(message)
                except Exception as exc:
                    span.record_exception(exc)

        # Handle callback appropriately from args or kwargs
        if "callback" in kwargs:
            callback = kwargs.get("callback")
            kwargs["callback"] = callback_with_instana
            return wrapped(*args, **kwargs)
        else:
            subscription, callback, *args = args
            args = (subscription, callback_with_instana, *args)
            return wrapped(*args, **kwargs)

    logger.debug("Instrumenting Google Cloud Pub/Sub")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instana ASGI Middleware
"""

from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict

from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace import SpanKind

from instana.log import logger
from instana.propagators.format import Format
from instana.singletons import agent, tracer
from instana.util.secrets import strip_secrets_from_query
from instana.util.traceutils import extract_custom_headers

if TYPE_CHECKING:
    from starlette.middleware.exceptions import ExceptionMiddleware
    from instana.span.span import InstanaSpan


class InstanaASGIMiddleware:
    """
    Instana ASGI Middleware
    """

    def __init__(self, app: "ExceptionMiddleware") -> None:
        self.app = app

    def _collect_kvs(self, scope: Dict[str, Any], span: "InstanaSpan") -> None:
        try:
            span.set_attribute("span.kind", SpanKind.SERVER)
            span.set_attribute("http.path", scope.get("path"))
            span.set_attribute(SpanAttributes.HTTP_METHOD, scope.get("method"))

            server = scope.get("server")
            if isinstance(server, tuple) or isinstance(server, list):
                span.set_attribute(SpanAttributes.HTTP_HOST, server[0])

            query = scope.get("query_string")
            if isinstance(query, (str, bytes)) and len(query):
                if isinstance(query, bytes):
                    query = query.decode("utf-8")
                scrubbed_params = strip_secrets_from_query(
                    query, agent.options.secrets_matcher, agent.options.secrets_list
                )
                span.set_attribute("http.params", scrubbed_params)

            app = scope.get("app")
            if app and hasattr(app, "routes"):
                # Attempt to detect the Starlette routes registered.
                # If Starlette isn't present, we harmlessly dump out.
                from starlette.routing import Match

                for route in scope["app"].routes:
                    if route.matches(scope)[0] == Match.FULL:
                        span.set_attribute("http.path_tpl", route.path)
        except Exception:
            logger.debug("ASGI collect_kvs: ", exc_info=True)

    async def __call__(
        self,
        scope: Dict[str, Any],
        receive: Callable[[], Awaitable[Dict[str, Any]]],
        send: Callable[[Dict[str, Any]], Awaitable[None]],
    ) -> None:
        request_context = None

        if scope["type"] not in ("http", "websocket"):
            return await self.app(scope, receive, send)

        request_headers = scope.get("headers")
        if isinstance(request_headers, list):
            request_context = tracer.extract(Format.BINARY, request_headers)

        with tracer.start_as_current_span("asgi", span_context=request_context) as span:
            self._collect_kvs(scope, span)
            if "headers" in scope:
                extract_custom_headers(span, scope["headers"])

            instana_send = self._send_with_instana(
                span,
                scope,
                send,
            )

            try:
                await self.app(scope, receive, instana_send)
            except Exception as exc:
                span.record_exception(exc)
                raise exc

    def _send_with_instana(
        self,
        current_span: "InstanaSpan",
        scope: Dict[str, Any],
        send: Callable[[Dict[str, Any]], Awaitable[None]],
    ) -> Awaitable[None]:
        async def send_wrapper(response: Dict[str, Any]) -> Awaitable[None]:
            if response["type"] == "http.response.start":
                try:
                    status_code = response.get("status")
                    if status_code:
                        if 500 <= int(status_code):
                            current_span.mark_as_errored()
                        current_span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, status_code)

                    headers = response.get("headers")
                    if headers:
                        extract_custom_headers(current_span, headers)
                        tracer.inject(current_span.context, Format.BINARY, headers)
                except Exception:
                    logger.debug("ASGI send_wrapper error: ", exc_info=True)

            try:
                await send(response)
            except Exception as exc:
                current_span.record_exception(exc)
                raise

        return send_wrapper



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2017


from typing import TYPE_CHECKING, Any, Callable, Dict, Tuple, Union

import wrapt
from opentelemetry.semconv.trace import SpanAttributes

from instana.log import logger
from instana.propagators.format import Format
from instana.singletons import agent
from instana.util.secrets import strip_secrets_from_query
from instana.util.traceutils import get_tracer_tuple, tracing_is_off, extract_custom_headers

if TYPE_CHECKING:
    from instana.span.span import InstanaSpan

try:
    import urllib3

    def _collect_kvs(
        instance: Union[
            urllib3.connectionpool.HTTPConnectionPool,
            urllib3.connectionpool.HTTPSConnectionPool,
        ],
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> Dict[str, Any]:
        kvs = dict()
        try:
            kvs["host"] = instance.host
            kvs["port"] = instance.port

            if args and len(args) == 2:
                kvs["method"] = args[0]
                kvs["path"] = args[1]
            else:
                kvs["method"] = kwargs.get("method")
                kvs["path"] = (
                    kwargs.get("path") if kwargs.get("path") else kwargs.get("url")
                )

            # Strip any secrets from potential query params
            if kvs.get("path") and ("?" in kvs["path"]):
                parts = kvs["path"].split("?")
                kvs["path"] = parts[0]
                if len(parts) == 2:
                    kvs["query"] = strip_secrets_from_query(
                        parts[1],
                        agent.options.secrets_matcher,
                        agent.options.secrets_list,
                    )

            url = kvs["host"] + ":" + str(kvs["port"]) + kvs["path"]
            if isinstance(instance, urllib3.connectionpool.HTTPSConnectionPool):
                kvs["url"] = f"https://{url}"
            else:
                kvs["url"] = f"http://{url}"
        except Exception:
            logger.debug("urllib3 _collect_kvs error: ", exc_info=True)
            return kvs
        else:
            return kvs

    def collect_response(
        span: "InstanaSpan", response: urllib3.response.HTTPResponse
    ) -> None:
        try:
            span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, response.status)

            extract_custom_headers(span, response.headers)

            if 500 <= response.status:
                span.mark_as_errored()
        except Exception:
            logger.debug("urllib3 collect_response error: ", exc_info=True)

    @wrapt.patch_function_wrapper("urllib3", "HTTPConnectionPool.urlopen")
    def urlopen_with_instana(
        wrapped: Callable[
            ..., Union[urllib3.HTTPConnectionPool, urllib3.HTTPSConnectionPool]
        ],
        instance: Union[
            urllib3.connectionpool.HTTPConnectionPool,
            urllib3.connectionpool.HTTPSConnectionPool,
        ],
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> urllib3.response.HTTPResponse:
        tracer, parent_span, span_name = get_tracer_tuple()

        # If we're not tracing, just return; boto3 has it's own visibility
        if tracing_is_off() or (span_name == "boto3"):
            return wrapped(*args, **kwargs)

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "urllib3", span_context=parent_context
        ) as span:
            try:
                kvs = _collect_kvs(instance, args, kwargs)
                if "url" in kvs:
                    span.set_attribute(SpanAttributes.HTTP_URL, kvs["url"])
                if "query" in kvs:
                    span.set_attribute("http.params", kvs["query"])
                if "method" in kvs:
                    span.set_attribute(SpanAttributes.HTTP_METHOD, kvs["method"])
                if "headers" in kwargs:
                    extract_custom_headers(span, kwargs["headers"])
                    tracer.inject(span.context, Format.HTTP_HEADERS, kwargs["headers"])

                response = wrapped(*args, **kwargs)

                collect_response(span, response)

                return response
            except Exception as e:
                span.record_exception(e)
                raise

    logger.debug("Instrumenting urllib3")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


from instana.log import logger
from instana.instrumentation.pep0249 import ConnectionFactory

try:
    import MySQLdb

    cf = ConnectionFactory(connect_func=MySQLdb.connect, module_name="mysql")

    setattr(MySQLdb, "connect", cf)
    if hasattr(MySQLdb, "Connect"):
        setattr(MySQLdb, "Connect", cf)

    logger.debug("Instrumenting mysqlclient")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2024

from typing import Generator, Tuple

import pytest

from instana.recorder import StanRecorder
from instana.span.sdk_span import SDKSpan
from instana.span.span import InstanaSpan
from instana.span_context import SpanContext


class TestSDKSpan:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.span = None
        yield

    def test_sdkspan(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-sdk-span"
        service_name = "test-sdk"
        attributes = {
            "span.kind": "entry",
            "arguments": "--quiet",
            "return": "True",
        }
        self.span = InstanaSpan(
            span_name, span_context, span_processor, attributes=attributes
        )
        sdk_span = SDKSpan(self.span, None, service_name)

        expected_result = {
            "n": "sdk",
            "k": 1,
            "data": {
                "service": service_name,
                "sdk": {
                    "name": span_name,
                    "type": attributes["span.kind"],
                    "custom": {
                        "attributes": attributes,
                    },
                    "arguments": attributes["arguments"],
                    "return": attributes["return"],
                },
            },
        }

        assert expected_result["n"] == sdk_span.n
        assert expected_result["k"] == sdk_span.k
        assert len(expected_result["data"]) == len(sdk_span.data)
        assert expected_result["data"]["service"] == sdk_span.data["service"]
        assert len(expected_result["data"]["sdk"]) == len(sdk_span.data["sdk"])
        assert expected_result["data"]["sdk"]["name"] == sdk_span.data["sdk"]["name"]
        assert expected_result["data"]["sdk"]["type"] == sdk_span.data["sdk"]["type"]
        assert len(attributes) == len(sdk_span.data["sdk"]["custom"]["tags"])
        assert attributes == sdk_span.data["sdk"]["custom"]["tags"]
        assert attributes["arguments"] == sdk_span.data["sdk"]["arguments"]
        assert attributes["return"] == sdk_span.data["sdk"]["return"]

    @pytest.mark.parametrize(
        "span_kind, expected_result",
        [
            (None, ("intermediate", 3)),
            ("entry", ("entry", 1)),
            ("server", ("entry", 1)),
            ("consumer", ("entry", 1)),
            ("exit", ("exit", 2)),
            ("client", ("exit", 2)),
            ("producer", ("exit", 2)),
        ],
    )
    def test_sdkspan_get_span_kind(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        span_kind: str,
        expected_result: Tuple[str, int],
    ) -> None:
        attributes = {
            "span.kind": span_kind,
        }
        self.span = InstanaSpan(
            "test-sdk-span", span_context, span_processor, attributes=attributes
        )
        sdk_span = SDKSpan(self.span, None, "test")

        kind = sdk_span.get_span_kind(self.span)

        assert expected_result == kind

    def test_sdkspan_get_span_kind_with_no_attributes(
        self,
        span: InstanaSpan,
    ) -> None:
        self.span = SDKSpan(span, None, "test")
        kind = self.span.get_span_kind(span)
        assert ("intermediate", 3) == kind



# (c) Copyright IBM Corp. 2024

from pytest import LogCaptureFixture

from instana.agent.base import BaseAgent
from instana.agent.host import HostAgent
from instana.propagators.binary_propagator import BinaryPropagator
from instana.propagators.format import Format
from instana.propagators.http_propagator import HTTPPropagator
from instana.propagators.text_propagator import TextPropagator
from instana.recorder import StanRecorder
from instana.sampling import InstanaSampler
from instana.tracer import InstanaTracer, InstanaTracerProvider


def test_tracer_provider_defaults() -> None:
    provider = InstanaTracerProvider()
    assert isinstance(provider.sampler, InstanaSampler)
    assert isinstance(provider._span_processor, StanRecorder)
    assert isinstance(provider._exporter, HostAgent)
    assert len(provider._propagators) == 3
    assert isinstance(provider._propagators[Format.HTTP_HEADERS], HTTPPropagator)
    assert isinstance(provider._propagators[Format.TEXT_MAP], TextPropagator)
    assert isinstance(provider._propagators[Format.BINARY], BinaryPropagator)


def test_tracer_provider_get_tracer() -> None:
    provider = InstanaTracerProvider()
    tracer = provider.get_tracer("instana.test.tracer")

    assert isinstance(tracer, InstanaTracer)


def test_tracer_provider_get_tracer_empty_instrumenting_module_name(
    caplog: LogCaptureFixture,
) -> None:
    provider = InstanaTracerProvider()
    tracer = provider.get_tracer("")

    assert "get_tracer called with missing module name." in caplog.messages
    assert isinstance(tracer, InstanaTracer)


def test_tracer_provider_add_span_processor(span_processor: StanRecorder) -> None:
    provider = InstanaTracerProvider()
    assert isinstance(provider._span_processor, StanRecorder)
    assert isinstance(provider._span_processor.agent, HostAgent)
    assert provider._span_processor.THREAD_NAME == "InstanaSpan Recorder"

    provider.add_span_processor(span_processor)
    assert isinstance(provider._span_processor, StanRecorder)
    assert isinstance(provider._span_processor.agent, BaseAgent)
    assert provider._span_processor.THREAD_NAME == "InstanaSpan Recorder Test"



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import time
import random
from typing import Generator

import pytest
import grpc

from opentelemetry.trace import SpanKind

import tests.apps.grpc_server  # noqa: F401
import tests.apps.grpc_server.stan_pb2 as stan_pb2
import tests.apps.grpc_server.stan_pb2_grpc as stan_pb2_grpc
from tests.helpers import testenv, get_first_span_by_name

from instana.singletons import tracer, agent
from instana.span.span import get_current_span


class TestGRPCIO:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Clear all spans before a test run"""
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        self.channel = grpc.insecure_channel(testenv["grpc_server"])
        self.server_stub = stan_pb2_grpc.StanStub(self.channel)
        # The grpc client apparently needs a second to connect and initialize
        time.sleep(1)
        yield
        # tearDown
        # Ensure that allow_exit_as_root has the default value
        agent.options.allow_exit_as_root = False

    def generate_questions(self) -> Generator[None, None, None]:
        """Used in the streaming grpc tests"""
        questions = [
            stan_pb2.QuestionRequest(question="Are you there?"),
            stan_pb2.QuestionRequest(question="What time is it?"),
            stan_pb2.QuestionRequest(question="Where in the world is Waldo?"),
            stan_pb2.QuestionRequest(
                question="What did one campfire say to the other?"
            ),
            stan_pb2.QuestionRequest(question="Is cereal soup?"),
            stan_pb2.QuestionRequest(
                question="What is always coming, but never arrives?"
            ),
        ]
        for q in questions:
            yield q
            time.sleep(random.uniform(0.2, 0.5))

    def test_vanilla_request(self) -> None:
        response = self.server_stub.OneQuestionOneResponse(
            stan_pb2.QuestionRequest(question="Are you there?")
        )
        assert (
            response.answer
            == "Invention, my dear friends, is 93% perspiration, 6% electricity, 4% evaporation, and 2% butterscotch ripple. – Willy Wonka"
        )

    def test_vanilla_request_via_with_call(self) -> None:
        response = self.server_stub.OneQuestionOneResponse.with_call(
            stan_pb2.QuestionRequest(question="Are you there?")
        )
        assert (
            response[0].answer
            == "Invention, my dear friends, is 93% perspiration, 6% electricity, 4% evaporation, and 2% butterscotch ripple. – Willy Wonka"
        )

    def test_unary_one_to_one(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.server_stub.OneQuestionOneResponse(
                stan_pb2.QuestionRequest(question="Are you there?")
            )

        assert not get_current_span().is_recording()
        assert response
        assert (
            response.answer
            == "Invention, my dear friends, is 93% perspiration, 6% electricity, 4% evaporation, and 2% butterscotch ripple. – Willy Wonka"
        )

        spans = self.recorder.queued_spans()

        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "rpc-server")
        client_span = get_first_span_by_name(spans, "rpc-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert server_span
        assert client_span
        assert test_span

        # Same traceId
        assert server_span.t == client_span.t
        assert server_span.t == test_span.t

        # Parent relationships
        assert server_span.p == client_span.s
        assert client_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        # rpc-server
        assert server_span.n == "rpc-server"
        assert server_span.k is SpanKind.SERVER
        assert not server_span.stack
        assert server_span.data["rpc"]["flavor"] == "grpc"
        assert server_span.data["rpc"]["call"] == "/stan.Stan/OneQuestionOneResponse"
        assert server_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert server_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert not server_span.data["rpc"]["error"]

        # rpc-client
        assert client_span.n == "rpc-client"
        assert client_span.k is SpanKind.CLIENT
        assert client_span.stack
        assert client_span.data["rpc"]["flavor"] == "grpc"
        assert client_span.data["rpc"]["call"] == "/stan.Stan/OneQuestionOneResponse"
        assert client_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert client_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert client_span.data["rpc"]["call_type"] == "unary"
        assert not client_span.data["rpc"]["error"]

        # test-span
        assert test_span.n == "sdk"
        assert test_span.data["sdk"]["name"] == "test"

    def test_streaming_many_to_one(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.server_stub.ManyQuestionsOneResponse(
                self.generate_questions()
            )

        assert not get_current_span().is_recording()
        assert response

        assert response.answer == "Ok"
        assert response.was_answered

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "rpc-server")
        client_span = get_first_span_by_name(spans, "rpc-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert server_span
        assert client_span
        assert test_span

        # Same traceId
        assert server_span.t == client_span.t
        assert server_span.t == test_span.t

        # Parent relationships
        assert server_span.p == client_span.s
        assert client_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        # rpc-server
        assert server_span.n == "rpc-server"
        assert server_span.k is SpanKind.SERVER
        assert not server_span.stack
        assert server_span.data["rpc"]["flavor"] == "grpc"
        assert server_span.data["rpc"]["call"] == "/stan.Stan/ManyQuestionsOneResponse"
        assert server_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert server_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert not server_span.data["rpc"]["error"]

        # rpc-client
        assert client_span.n == "rpc-client"
        assert client_span.k is SpanKind.CLIENT
        assert client_span.stack
        assert client_span.data["rpc"]["flavor"] == "grpc"
        assert client_span.data["rpc"]["call"] == "/stan.Stan/ManyQuestionsOneResponse"
        assert client_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert client_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert client_span.data["rpc"]["call_type"] == "stream"
        assert not client_span.data["rpc"]["error"]

        # test-span
        assert test_span.n == "sdk"
        assert test_span.data["sdk"]["name"] == "test"

    def test_streaming_one_to_many(self) -> None:
        with tracer.start_as_current_span("test"):
            responses = self.server_stub.OneQuestionManyResponses(
                stan_pb2.QuestionRequest(question="Are you there?")
            )

        assert not get_current_span().is_recording()
        assert responses

        final_answers = []
        for response in responses:
            final_answers.append(response)

        assert len(final_answers) == 6

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "rpc-server")
        client_span = get_first_span_by_name(spans, "rpc-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert server_span
        assert client_span
        assert test_span

        # Same traceId
        assert server_span.t == client_span.t
        assert server_span.t == test_span.t

        # Parent relationships
        assert server_span.p == client_span.s
        assert client_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        # rpc-server
        assert server_span.n == "rpc-server"
        assert server_span.k is SpanKind.SERVER
        assert not server_span.stack
        assert server_span.data["rpc"]["flavor"] == "grpc"
        assert server_span.data["rpc"]["call"] == "/stan.Stan/OneQuestionManyResponses"
        assert server_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert server_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert not server_span.data["rpc"]["error"]

        # rpc-client
        assert client_span.n == "rpc-client"
        assert client_span.k is SpanKind.CLIENT
        assert client_span.stack
        assert client_span.data["rpc"]["flavor"] == "grpc"
        assert client_span.data["rpc"]["call"] == "/stan.Stan/OneQuestionManyResponses"
        assert client_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert client_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert client_span.data["rpc"]["call_type"] == "stream"
        assert not client_span.data["rpc"]["error"]

        # test-span
        assert test_span.n == "sdk"
        assert test_span.data["sdk"]["name"] == "test"

    def test_streaming_many_to_many(self) -> None:
        with tracer.start_as_current_span("test"):
            responses = self.server_stub.ManyQuestionsManyReponses(
                self.generate_questions()
            )

        assert not get_current_span().is_recording()
        assert responses

        final_answers = []
        for response in responses:
            final_answers.append(response)

        assert len(final_answers) == 6

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "rpc-server")
        client_span = get_first_span_by_name(spans, "rpc-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert server_span
        assert client_span
        assert test_span

        # Same traceId
        assert server_span.t == client_span.t
        assert server_span.t == test_span.t

        # Parent relationships
        assert server_span.p == client_span.s
        assert client_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        # rpc-server
        assert server_span.n == "rpc-server"
        assert server_span.k is SpanKind.SERVER
        assert not server_span.stack
        assert server_span.data["rpc"]["flavor"] == "grpc"
        assert server_span.data["rpc"]["call"] == "/stan.Stan/ManyQuestionsManyReponses"
        assert server_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert server_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert not server_span.data["rpc"]["error"]

        # rpc-client
        assert client_span.n == "rpc-client"
        assert client_span.k is SpanKind.CLIENT
        assert client_span.stack
        assert client_span.data["rpc"]["flavor"] == "grpc"
        assert client_span.data["rpc"]["call"] == "/stan.Stan/ManyQuestionsManyReponses"
        assert client_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert client_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert client_span.data["rpc"]["call_type"] == "stream"
        assert not client_span.data["rpc"]["error"]

        # test-span
        assert test_span.n == "sdk"
        assert test_span.data["sdk"]["name"] == "test"

    def test_unary_one_to_one_with_call(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.server_stub.OneQuestionOneResponse.with_call(
                stan_pb2.QuestionRequest(question="Are you there?")
            )

        assert not get_current_span().is_recording()
        assert response
        assert isinstance(response, tuple)
        assert (
            response[0].answer
            == "Invention, my dear friends, is 93% perspiration, 6% electricity, 4% evaporation, and 2% butterscotch ripple. – Willy Wonka"
        )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "rpc-server")
        client_span = get_first_span_by_name(spans, "rpc-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert server_span
        assert client_span
        assert test_span

        # Same traceId
        assert server_span.t == client_span.t
        assert server_span.t == test_span.t

        # Parent relationships
        assert server_span.p == client_span.s
        assert client_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        # rpc-server
        assert server_span.n == "rpc-server"
        assert server_span.k is SpanKind.SERVER
        assert not server_span.stack
        assert server_span.data["rpc"]["flavor"] == "grpc"
        assert server_span.data["rpc"]["call"] == "/stan.Stan/OneQuestionOneResponse"
        assert server_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert server_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert not server_span.data["rpc"]["error"]

        # rpc-client
        assert client_span.n == "rpc-client"
        assert client_span.k is SpanKind.CLIENT
        assert client_span.stack
        assert client_span.data["rpc"]["flavor"] == "grpc"
        assert client_span.data["rpc"]["call"] == "/stan.Stan/OneQuestionOneResponse"
        assert client_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert client_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert client_span.data["rpc"]["call_type"] == "unary"
        assert not client_span.data["rpc"]["error"]

        # test-span
        assert test_span.n == "sdk"
        assert test_span.data["sdk"]["name"] == "test"

    def test_streaming_many_to_one_with_call(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.server_stub.ManyQuestionsOneResponse.with_call(
                self.generate_questions()
            )

        assert not get_current_span().is_recording()
        assert response

        assert response[0].answer == "Ok"
        assert response[0].was_answered

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "rpc-server")
        client_span = get_first_span_by_name(spans, "rpc-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert server_span
        assert client_span
        assert test_span

        # Same traceId
        assert server_span.t == client_span.t
        assert server_span.t == test_span.t

        # Parent relationships
        assert server_span.p == client_span.s
        assert client_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        # rpc-server
        assert server_span.n == "rpc-server"
        assert server_span.k is SpanKind.SERVER
        assert not server_span.stack
        assert server_span.data["rpc"]["flavor"] == "grpc"
        assert server_span.data["rpc"]["call"] == "/stan.Stan/ManyQuestionsOneResponse"
        assert server_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert server_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert not server_span.data["rpc"]["error"]

        # rpc-client
        assert client_span.n == "rpc-client"
        assert client_span.k is SpanKind.CLIENT
        assert client_span.stack
        assert client_span.data["rpc"]["flavor"] == "grpc"
        assert client_span.data["rpc"]["call"] == "/stan.Stan/ManyQuestionsOneResponse"
        assert client_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert client_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert client_span.data["rpc"]["call_type"] == "stream"
        assert not client_span.data["rpc"]["error"]

        # test-span
        assert test_span.n == "sdk"
        assert test_span.data["sdk"]["name"] == "test"

    def test_async_unary(self) -> None:
        def process_response(future):
            result = future.result()
            assert isinstance(result, stan_pb2.QuestionResponse)
            assert result.was_answered
            assert (
                result.answer
                == "Invention, my dear friends, is 93% perspiration, 6% electricity, 4% evaporation, and 2% butterscotch ripple. – Willy Wonka"
            )

        with tracer.start_as_current_span("test"):
            future = self.server_stub.OneQuestionOneResponse.future(
                stan_pb2.QuestionRequest(question="Are you there?")
            )
            future.add_done_callback(process_response)
            time.sleep(0.7)

        assert not get_current_span().is_recording()
        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "rpc-server")
        client_span = get_first_span_by_name(spans, "rpc-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert server_span
        assert client_span
        assert test_span

        # Same traceId
        assert server_span.t == client_span.t
        assert server_span.t == test_span.t

        # Parent relationships
        assert server_span.p == client_span.s
        assert client_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        # rpc-server
        assert server_span.n == "rpc-server"
        assert server_span.k is SpanKind.SERVER
        assert not server_span.stack
        assert server_span.data["rpc"]["flavor"] == "grpc"
        assert server_span.data["rpc"]["call"] == "/stan.Stan/OneQuestionOneResponse"
        assert server_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert server_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert not server_span.data["rpc"]["error"]

        # rpc-client
        assert client_span.n == "rpc-client"
        assert client_span.k is SpanKind.CLIENT
        assert client_span.stack
        assert client_span.data["rpc"]["flavor"] == "grpc"
        assert client_span.data["rpc"]["call"] == "/stan.Stan/OneQuestionOneResponse"
        assert client_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert client_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert client_span.data["rpc"]["call_type"] == "unary"
        assert not client_span.data["rpc"]["error"]

        # test-span
        assert test_span.n == "sdk"
        assert test_span.data["sdk"]["name"] == "test"

    def test_async_stream(self) -> None:
        def process_response(future):
            result = future.result()
            assert isinstance(result, stan_pb2.QuestionResponse)
            assert result.was_answered
            assert result.answer == "Ok"

        with tracer.start_as_current_span("test"):
            future = self.server_stub.ManyQuestionsOneResponse.future(
                self.generate_questions()
            )
            future.add_done_callback(process_response)

        # The question generator delays at random intervals between questions so to assure that
        # all questions are sent and processed before we start testing the results.
        time.sleep(5)

        assert not get_current_span().is_recording()
        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        server_span = get_first_span_by_name(spans, "rpc-server")
        client_span = get_first_span_by_name(spans, "rpc-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert server_span
        assert client_span
        assert test_span

        # Same traceId
        assert server_span.t == client_span.t
        assert server_span.t == test_span.t

        # Parent relationships
        assert server_span.p == client_span.s
        assert client_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not client_span.ec
        assert not server_span.ec

        # rpc-server
        assert server_span.n == "rpc-server"
        assert server_span.k is SpanKind.SERVER
        assert not server_span.stack
        assert server_span.data["rpc"]["flavor"] == "grpc"
        assert server_span.data["rpc"]["call"] == "/stan.Stan/ManyQuestionsOneResponse"
        assert server_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert server_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert not server_span.data["rpc"]["error"]

        # rpc-client
        assert client_span.n == "rpc-client"
        assert client_span.k is SpanKind.CLIENT
        assert client_span.stack
        assert client_span.data["rpc"]["flavor"] == "grpc"
        assert client_span.data["rpc"]["call"] == "/stan.Stan/ManyQuestionsOneResponse"
        assert client_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert client_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert client_span.data["rpc"]["call_type"] == "stream"
        assert not client_span.data["rpc"]["error"]

        # test-span
        assert test_span.n == "sdk"
        assert test_span.data["sdk"]["name"] == "test"

    def test_server_error(self) -> None:
        response = None
        with tracer.start_as_current_span("test"):
            try:
                response = self.server_stub.OneQuestionOneErrorResponse(
                    stan_pb2.QuestionRequest(question="Do u error?")
                )
            except Exception:
                pass

        assert not get_current_span().is_recording()
        assert not response

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        log_span = get_first_span_by_name(spans, "log")
        server_span = get_first_span_by_name(spans, "rpc-server")
        client_span = get_first_span_by_name(spans, "rpc-client")
        test_span = get_first_span_by_name(spans, "sdk")

        assert log_span
        assert server_span
        assert client_span
        assert test_span

        # Same traceId
        assert server_span.t == client_span.t
        assert server_span.t == test_span.t

        # Parent relationships
        assert server_span.p == client_span.s
        assert client_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert client_span.ec == 1
        assert not server_span.ec

        # rpc-server
        assert server_span.n == "rpc-server"
        assert server_span.k is SpanKind.SERVER
        assert not server_span.stack
        assert server_span.data["rpc"]["flavor"] == "grpc"
        assert (
            server_span.data["rpc"]["call"] == "/stan.Stan/OneQuestionOneErrorResponse"
        )
        assert server_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert server_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert not server_span.data["rpc"]["error"]

        # rpc-client
        assert client_span.n == "rpc-client"
        assert client_span.k is SpanKind.CLIENT
        assert client_span.stack
        assert client_span.data["rpc"]["flavor"] == "grpc"
        assert (
            client_span.data["rpc"]["call"] == "/stan.Stan/OneQuestionOneErrorResponse"
        )
        assert client_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert client_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert client_span.data["rpc"]["call_type"] == "unary"
        assert client_span.data["rpc"]["error"]

        # log
        assert log_span.n == "log"
        assert log_span.data["log"]
        assert (
            log_span.data["log"]["message"]
            == "Exception calling application: Simulated error"
        )

        # test-span
        assert test_span.n == "sdk"
        assert test_span.data["sdk"]["name"] == "test"

    def test_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True

        response = self.server_stub.OneQuestionOneResponse.with_call(
            stan_pb2.QuestionRequest(question="Are you there?")
        )
        assert (
            response[0].answer
            == "Invention, my dear friends, is 93% perspiration, 6% electricity, 4% evaporation, and 2% butterscotch ripple. – Willy Wonka"
        )

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        server_span = spans[0]

        assert server_span

        # Parent relationships
        assert not server_span.p

        # Error logging
        assert not server_span.ec

        # rpc-server
        assert server_span.n == "rpc-server"
        assert server_span.k is SpanKind.SERVER
        assert not server_span.stack
        assert server_span.data["rpc"]["flavor"] == "grpc"
        assert server_span.data["rpc"]["call"] == "/stan.Stan/OneQuestionOneResponse"
        assert server_span.data["rpc"]["host"] == testenv["grpc_host"]
        assert server_span.data["rpc"]["port"] == str(testenv["grpc_port"])
        assert not server_span.data["rpc"]["error"]

    def test_no_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = False
        responses = self.server_stub.OneQuestionManyResponses(
            stan_pb2.QuestionRequest(question="Are you there?")
        )

        assert responses

        spans = self.recorder.queued_spans()
        assert len(spans) == 0



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

import unittest

from instana.configurator import config


class TestRedis(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_has_default_config(self):
        self.assertEqual(config['asyncio_task_context_propagation']['enabled'], False)


# (c) Copyright IBM Corp. 2024

import pytest
from opentelemetry.trace.span import _SPAN_ID_MAX_VALUE

from instana.agent.host import HostAgent
from instana.recorder import StanRecorder
from instana.sampling import InstanaSampler
from instana.span.span import (
    INVALID_SPAN,
    INVALID_SPAN_ID,
    InstanaSpan,
    get_current_span,
)
from instana.span_context import SpanContext
from instana.tracer import InstanaTracer, InstanaTracerProvider


def test_tracer_defaults(tracer_provider: InstanaTracerProvider) -> None:
    tracer = InstanaTracer(
        tracer_provider.sampler,
        tracer_provider._span_processor,
        tracer_provider._exporter,
        tracer_provider._propagators,
    )

    assert isinstance(tracer._sampler, InstanaSampler)
    assert isinstance(tracer.span_processor, StanRecorder)
    assert isinstance(tracer.exporter, HostAgent)
    assert len(tracer._propagators) == 3


def test_tracer_start_span(
    tracer_provider: InstanaTracerProvider, span_context: SpanContext
) -> None:
    span_name = "test-span"
    tracer = InstanaTracer(
        tracer_provider.sampler,
        tracer_provider._span_processor,
        tracer_provider._exporter,
        tracer_provider._propagators,
    )
    span = tracer.start_span(name=span_name, span_context=span_context)

    assert span
    assert isinstance(span, InstanaSpan)
    assert span.name == span_name
    assert not span.stack


def test_tracer_start_span_with_stack(tracer_provider: InstanaTracerProvider) -> None:
    span_name = "log"
    tracer = InstanaTracer(
        tracer_provider.sampler,
        tracer_provider._span_processor,
        tracer_provider._exporter,
        tracer_provider._propagators,
    )
    span = tracer.start_span(name=span_name)

    assert span
    assert isinstance(span, InstanaSpan)
    assert span.name == span_name
    assert span.stack

    stack_0 = span.stack[0]
    assert 3 == len(stack_0)
    assert "c" in stack_0.keys()
    assert "n" in stack_0.keys()
    assert "m" in stack_0.keys()


def test_tracer_start_span_Exception(
    mocker, tracer_provider: InstanaTracerProvider, span_context: SpanContext
) -> None:
    span_name = "test-span"
    tracer = InstanaTracer(
        tracer_provider.sampler,
        tracer_provider._span_processor,
        tracer_provider._exporter,
        tracer_provider._propagators,
    )

    mocker.patch(
        "instana.tracer.InstanaTracer._create_span_context",
        return_value={"key": "value"},
    )
    with pytest.raises(AttributeError):
        tracer.start_span(name=span_name, span_context=span_context)


def test_tracer_start_as_current_span(tracer_provider: InstanaTracerProvider) -> None:
    span_name = "test-span"
    tracer = InstanaTracer(
        tracer_provider.sampler,
        tracer_provider._span_processor,
        tracer_provider._exporter,
        tracer_provider._propagators,
    )
    with tracer.start_as_current_span(name=span_name) as span:
        assert span is not None
        assert isinstance(span, InstanaSpan)
        assert span.name == span_name


def test_tracer_nested_span(tracer_provider: InstanaTracerProvider) -> None:
    tracer = InstanaTracer(
        tracer_provider.sampler,
        tracer_provider._span_processor,
        tracer_provider._exporter,
        tracer_provider._propagators,
    )
    parent_span_name = "parent-span"
    child_span_name = "child-span"
    with tracer.start_as_current_span(name=parent_span_name) as pspan:
        assert get_current_span() is pspan
        with tracer.start_as_current_span(name=child_span_name) as cspan:
            assert get_current_span() is cspan
            assert cspan.parent_id == pspan.context.span_id
        # child span goes out of scope
        assert cspan.end_time is not None
        assert get_current_span() is pspan
    # parent span goes out of scope
    assert pspan.end_time is not None
    assert get_current_span() is INVALID_SPAN


def test_tracer_create_span_context(
    span_context: SpanContext, tracer_provider: InstanaTracerProvider
) -> None:
    tracer = InstanaTracer(
        tracer_provider.sampler,
        tracer_provider._span_processor,
        tracer_provider._exporter,
        tracer_provider._propagators,
    )
    new_span_context = tracer._create_span_context(span_context)

    assert span_context.trace_id == new_span_context.trace_id
    assert span_context.span_id != new_span_context.span_id
    assert span_context.long_trace_id == new_span_context.long_trace_id

    assert span_context.trace_id > INVALID_SPAN_ID
    assert span_context.trace_id <= _SPAN_ID_MAX_VALUE

    assert span_context.span_id > INVALID_SPAN_ID
    assert span_context.span_id <= _SPAN_ID_MAX_VALUE


def test_tracer_create_span_context_root(
    tracer_provider: InstanaTracerProvider,
) -> None:
    tracer = InstanaTracer(
        tracer_provider.sampler,
        tracer_provider._span_processor,
        tracer_provider._exporter,
        tracer_provider._propagators,
    )
    new_span_context = tracer._create_span_context(parent_context=None)

    assert new_span_context.trace_id > INVALID_SPAN_ID
    assert new_span_context.trace_id <= _SPAN_ID_MAX_VALUE

    assert new_span_context.trace_id == new_span_context.span_id


def test_tracer_add_stack_high_limit(
    span: InstanaSpan, tracer_provider: InstanaTracerProvider
) -> None:
    tracer = InstanaTracer(
        tracer_provider.sampler,
        tracer_provider._span_processor,
        tracer_provider._exporter,
        tracer_provider._propagators,
    )
    tracer._add_stack(span, 50)

    assert span.stack
    assert 40 >= len(span.stack)

    stack_0 = span.stack[0]
    assert 3 == len(stack_0)
    assert "c" in stack_0.keys()
    assert "n" in stack_0.keys()
    assert "m" in stack_0.keys()


def test_tracer_add_stack_low_limit(
    span: InstanaSpan, tracer_provider: InstanaTracerProvider
) -> None:
    tracer = InstanaTracer(
        tracer_provider.sampler,
        tracer_provider._span_processor,
        tracer_provider._exporter,
        tracer_provider._propagators,
    )
    tracer._add_stack(span, 5)

    assert span.stack
    assert 5 >= len(span.stack)

    stack_0 = span.stack[0]
    assert 3 == len(stack_0)
    assert "c" in stack_0.keys()
    assert "n" in stack_0.keys()
    assert "m" in stack_0.keys()
